/*
 * pi - P user Interface
 * 
 * $Id: tag.c,v 1.1.1.1 2001/02/13 18:30:27 malekith Exp $
 * Author: Michal Moskal <malekith@pld.org.pl>
 * include COPYING-GNU
 */

#include <pi/pi.h>
#include <stdlib.h>

int pi_exists(struct pi_object *ctx, unsigned long tag)
{
	struct pi_tag *p, *e = 0;

	p = ctx->tags;
	for (;;) {
		if (p->tag == 0) {
			if ((p = e) == 0)
				return 0;
			e = 0;
		}

		if (pi_tag_id(p->tag) == tag)
			return 1;
		if (p->tag == pi_tag_super)
			e = ((struct pi_object*)p->data)->tags;
		p++;
	}
}

void *pi_fetch_data(struct pi_object *ctx, unsigned long tag, void *def)
{
	struct pi_tag *p;
	struct pi_object *super = ctx;

restart_with_super:
	if (super == 0)
		return def;
		
	p = super->tags;
	super = 0;
	
	while (p->tag) {
		if (pi_tag_id(p->tag) == tag)
			break;
		if (p->tag == pi_tag_super)
			super = p->data;
		p++;
	}

	if (p->tag == 0) 
		goto restart_with_super;
		
	return (p->tag & pi_tf_func) ?
		((void *(*)(struct pi_object *))p->data)(ctx) :
		p->data;
}

void pi_set_tag(struct pi_object *ctx, unsigned long tag, ...)
{
	va_list ap;
	struct pi_tag *p;
	void *ptr;

	va_start(ap, tag);
	ptr = va_arg(ap, void*);
	va_end(ap);

	p = ctx->tags;
	
	while (p->tag) {
		if (pi_tag_id(p->tag) == tag)
			break;
		p++;
	}

	if (p->tag == 0) 
		pi_add_object(ctx, tag, ptr, 0);
	else {
		/* this is for the case when flags are
		 * changed 
		 */
		p->tag = tag;
		p->data = ptr;
	}
}

struct pi_object *pi_add_objectv(struct pi_object *ctx, va_list ap)
{
	unsigned long tag;
	void *data;

	if (ctx == 0) {
		ctx = calloc(1, sizeof(struct pi_object));
		if (ctx == 0)
			pi_die("out of memory");
		ctx->used = 1;
		ctx->tags = calloc(ctx->allocated = 8, 
				   sizeof(struct pi_tag));
		if (ctx->tags == 0)
			pi_die("out of memory");
		ctx->tags[0].tag = 0;
		ctx->tags[0].data = 0;
	}
	
	while ((tag = va_arg(ap, unsigned long))) {
		data = va_arg(ap, void*);
		if (ctx->used == ctx->allocated) {
			ctx->tags = realloc(ctx->tags, 
			   (ctx->allocated *= 2) * sizeof(struct pi_tag));
			if (ctx->tags == 0)
				pi_die("out of memory");
		}
		ctx->tags[ctx->used-1].tag = tag;
		ctx->tags[ctx->used-1].data = data;
		ctx->tags[ctx->used].tag = 0;
		ctx->tags[ctx->used].data = 0;
		ctx->used++;
	}

	return ctx;
}

struct pi_object *pi_add_object(struct pi_object *ctx, ... )
{
	va_list ap;
	struct pi_object *r;
	
	va_start(ap, ctx);
	r = pi_add_objectv(ctx, ap);
	va_end(ap);

	return r;
}

void pi_kill_object(struct pi_object *ctx)
{
	struct pi_tag *p;

	for (p = ctx->tags; p->tag; p++)
		if (p->tag & pi_tf_free)
			free(p->data);

	free(ctx->tags);
	free(ctx);
}
